#define __KERNEL__
#define MODULE
#define trace_print(format, args...) printk(format, ## args)

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <sys/syscall.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <sys/poll.h>
#include <linux/dirent.h>
#include <utime.h>
//#include <linux/time.h>

extern void* sys_call_table[];

/* --------------------------- definitionen ---------------------------	*/

int (*orig_access) (const char *path, int mode);
int (*orig_bdflush) (int func, ...);
int (*orig_chdir) (const char *path);
int (*orig_chmod) (const char *path, mode_t mode);
int (*orig_chown) (const char *path, uid_t owner, gid_t group);
int (*orig_chown32) (const char *path, uid_t owner, gid_t group);
int (*orig_chroot) (const char *path);
int (*orig_close) (int fd);
int (*orig_creat) (const char *path, mode_t mode);
int (*orig_dup) (int oldfd);
int (*orig_dup2) (int oldfd, int newfd);
int (*orig_execve) (const char *path, char *const argv[], char *const envp[]);
int (*orig_fchdir) (int fd);
int (*orig_fchmod) (int fd, mode_t mode);
int (*orig_fchown) (int fd, uid_t owner, gid_t group);
int (*orig_fchown32) (int fd, uid_t owner, gid_t group);
int (*orig_fcntl) (int fd, int cmd, ...);
int (*orig_fcntl64) (int fd, int cmd, ...);

#ifdef _POSIX_SYNCHRONIZED_IO
int (*orig_fdatasync) (int fd);
#endif

int (*orig_flock) (int fd, int operation);
int (*orig_fstat) (int fd, struct stat *buf);
int (*orig_fstat64) (int fd, struct stat *buf);
int (*orig_fstatfs) (int fd, struct statfs *buf);
int (*orig_fsync) (int fd);
int (*orig_ftruncate) (int fd, off_t length);
int (*orig_ftruncate64) (int fd, off_t length);
int (*orig_getdents) (unsigned int fd, struct dirent *dirp, unsigned int count);
int (*orig_getdents64) (unsigned int fd, struct dirent *dirp, unsigned int count);
int (*orig_ioctl) (int fd, int request, ...);
int (*orig_lchown) (const char *path, uid_t owner, gid_t group);
int (*orig_lchown32) (const char *path, uid_t owner, gid_t group);
int (*orig_link) (const char *oldpath, const char *newpath);
int (*orig__llseek) (unsigned int fd, unsigned long offset_high, unsigned long offset_low, loff_t *result, unsigned int whence);
off_t (*orig_lseek) (int fd, off_t offset, int whence);
int (*orig_lstat) (const char *fd, struct stat *buf);
int (*orig_lstat64) (const char *fd, struct stat *buf);
int (*orig_mkdir) (const char *path, mode_t mode);
int (*orig_mknod) (const char *path, mode_t mode, dev_t dev);
int (*orig_mount) (const char *source, const char *target, const char *filesystemtype, unsigned long mountflags, const void *data);
int (*orig_open) (const char *path, int flags, ...);
int (*orig_pipe) (int fd[2]);
int (*orig_pivot_root) (const char *new_root, const char *put_old);
int (*orig_poll) (struct pollfd *ufds, unsigned int nfds, int timeout);
ssize_t (*orig_pread) (int fd, void *buf, size_t count, off_t offset);
ssize_t (*orig_pwrite) (int fd, const void *buf, size_t count, off_t offset);
ssize_t (*orig_read) (int fd, void *buf, size_t count);
int (*orig_readdir) (unsigned int fd, struct dirent *dirp, unsigned int count);
int (*orig_readlink) (const char *path, char *buf, size_t bufsiz);
int (*orig_readv) (int fd, const struct iovec *vector, size_t count);
int (*orig_rename) (const char *oldpath, const char *newpath);
int (*orig_rmdir) (const char *pathname);
int (*orig_select) (int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout);
int (*orig_stat) (const char *path, struct stat *buf);
int (*orig_stat64) (const char *path, struct stat *buf);
int (*orig_statfs) (const char *path, struct statfs *buf);
int (*orig_symlink) (const char *oldpath, const char *newpath);
void (*orig_sync) (void);
int (*orig_sysfs) (int option, ... );
int (*orig_truncate) (const char *path, off_t length);
int (*orig_truncate64) (const char *path, off_t length);
mode_t (*orig_umask) (mode_t mask);
int (*orig_umount) (const char *target);
int (*orig_unlink) (const char *pathname);
int (*orig_uselib) (const char *library);
int (*orig_ustat) (dev_t dev, struct ustat *ubuf);
int (*orig_utime) (const char *path, struct utimbuf *buf);
int (*orig_vhangup) (void);
ssize_t (*orig_write) (int fd, const void *buf, size_t count);
int (*orig_writev) (int fd, const struct iovec *vector, size_t count);

void print (char *format,...)
{
  void *args = __builtin_apply_args();
  void *ret = __builtin_apply((void *) printk, args, 128);
  __builtin_return(ret);
}//print


//time_t start;

/* --------------------------- Trace Funkionen --------------------------- */

struct modes {
  int value;
  char *name;
};

int first = 0;

//systemcall open parameter flags
struct modes fcntl[] = {
			{O_RDONLY,    "O_RDONLY"},     //00
	                {O_WRONLY,    "O_WRONLY"},     //01
                   	{O_RDWR,      "O_RDWR"  },     //02
                   	{O_CREAT,     "O_CREAT" },     //100
                   	{O_EXCL,      "O_EXCL"  },     //200
                   	{O_NOCTTY,    "O_NOCTTY"},     //400
                   	{O_TRUNC,     "O_TRUNC"  },    //1000
                   	{O_APPEND,    "O_APPEN" },     //2000
                   	{O_NONBLOCK,  "O_NONBLOCK"},   //4000
                      //{O_NDELAY,    "O_NDELAY"},     //4000
                   	{O_SYNC,      "O_SYNC"  },     //10000
                   	{O_NOFOLLOW,  "O_NOFOLLOW"},   //400000
                   	{O_DIRECTORY, "O_DIRECTORY"},  //200000
                   	{O_LARGEFILE, "O_LARGEFILE"},  //100000
                   	{0,    NULL}
		      };


int trace_open(const char *path, int flags, ...)
{
  char *ok = "failed";
  int n=0, i=0;
  
  void *args = __builtin_apply_args();
  int *result = (int *) __builtin_apply ((void *) orig_open, args, 128);
  if(*result != -1)
    ok = "done";

  if(flags == 0 && fcntl[i].value == 0) {
	trace_print("<1>vfs_trace: open (\"%s\", %s) ... %s\n", path, fcntl->name, ok);
  }//if
  else {
	  trace_print("<1>vfs_trace: open (\"%s\", ", path);
	  for(; fcntl[i].name && flags; i++) {
	    if(fcntl[i].value && (fcntl[i].value & flags) == fcntl[i].value) {
		if(n>0)
		   trace_print("|");
		trace_print("%s", fcntl[i].name);
		flags &= ~fcntl[i].value;
		n++;
	    }//if
	  }//for
 	  trace_print(") ... %s\n", ok);
  }//else

  __builtin_return (result);
}//trace_open

struct modes access_mode[] = {
				{4, "R_OK"},
				{2, "W_OK"},
				{1, "X_OK"},
				{0, "F_OK"},
				{0, NULL},
			     };

int trace_access (const char *path, int mode)
{
  char *ok = "failed"; 
  int ret = (*orig_access) (path,mode);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: access (%s) ... %s\n", path, ok);
  return ret;
}//trace_access

int trace_bdflush (int func, ...)
{
  char *ok = "failed";
  void *args;
  int *result;
  args = __builtin_apply_args();
  result = (int *) __builtin_apply ((void *) orig_fcntl,args, 128);
  if(*result > -1)
    ok = "done";
  trace_print("<1>vfs_trace: bdflush (%d) ... %s",func,ok);
  __builtin_return (result);
}//trace_bdflush

int trace_chdir (const char *path)
{
  char *ok = "failed";
  int ret = (*orig_chdir) (path);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: chdir (%s) ... %s\n", path, ok);
  return ret;
}//trace_chdir

struct modes chmod_mode[] = {
				{S_ISUID, "S_ISUID"},
				{S_ISGID, "S_ISGID"},
				{S_ISVTX, "S_ISVTX"},
				{S_IRUSR, "S_IRUSR"},
				{S_IWUSR, "S_IWUSR"},
				{S_IXUSR, "S_IXUSR"},
				{S_IRGRP, "S_IRGRP"},
				{S_IWGRP, "S_IWGRP"},
				{S_IXGRP, "S_IXGRP"},
				{S_IROTH, "S_IROTH"},
				{S_IWOTH, "S_IWOTH"},
				{S_IXOTH, "S_IXOTH"},
				{0,	  NULL},
			     };

int trace_chmod (const char *path, mode_t mode)
{
  char *ok = "failed";
  int ret = (*orig_chmod) (path, mode);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: chmod (%s) ... %s\n", path, ok);
  return ret;
}//trace_chmod

int trace_chown (const char *path, uid_t owner, gid_t group)
{
  char *ok = "failed";
  int ret = (*orig_chown) (path, owner, group);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: chown (%s,%d,%d) ... %s\n", path, owner, group, ok);
  return ret;
}//trace_chown

int trace_chown32 (const char *path, uid_t owner, gid_t group)
{
  char *ok = "failed";
  int ret = (*orig_chown32) (path, owner, group);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: chown32 (%s,%d,%d) ... %s\n", path, owner, group, ok);
  return ret;
}//trace_chown64

int trace_chroot (const char *path)
{
  char *ok = "failed";
  int ret = (*orig_chroot) (path);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: chroot (%s) ... %s\n", path, ok);
  return ret;
}//trace_chroot

int trace_close (int fd)
{
  char *ok = "failed";  
  int ret = (*orig_close) (fd);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: close (%d) ... %s\n",fd,ok);
  return ret;
}//trace_close

int trace_creat (const char *path, mode_t mode)
{
  char *ok = "failed";
  int ret = (*orig_creat) (path,mode);
  if(ret > -1)
    ok = "done";
  trace_print("<1>vfs_trace: creat (%s) ... %s\n", path, ok);
  return ret;
}//trace_creat

int trace_dup (int oldfd)
{
  char *ok = "failed", *path;
  int ret = (*orig_dup) (oldfd);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[oldfd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: dup (%s) ... %s\n", path, ok);
  return ret;
}//trace_dup

int trace_dup2 (int oldfd, int newfd)
{
  char *ok = "failed";
  int ret = (*orig_dup2) (oldfd, newfd);
  char *oldpath = (char *) current->files->fd[oldfd]->f_dentry->d_name.name;
  char *newpath = (char *) current->files->fd[newfd]->f_dentry->d_name.name;
  if(ret > -1)
    ok = "done";
  trace_print("<1>vfs_trace: dup2 (%s,%s) ... %s\n", oldpath, newpath, ok);
  return ret;
}//trace_dup2

int trace_execve (const char *path, char *const argv[], char *const envp[])
{
  char *ok = "failed";
  int ret;   
  ret = (*orig_execve) (path,argv,envp);
  if(ret != -1)
    ok = "done";
  trace_print("<1>vfs_trace: execve (%s) ... %s\n", path, ok);
  return ret;
}//trace_execve

#ifdef _POSIX_SYNCHRONIZED_IO
int trace_fdatasync (int fd)
{
  char *ok = "failed";
  int ret;
  ret = (*orig_fdatasync) (fd);
  if(ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fdatasync (%s) ... %s\n", path, ok);
  return ret;
}//trace_fdatasync
#endif

int trace_fchdir (int fd)
{
  char *ok = "failed", *path;
  int ret = (*orig_fchdir) (fd);
  if(ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fchdir (%s) ... %s\n", path, ok);
  return ret;
}//trace_fchdir

int trace_fchmod (int fd, mode_t mode)
{
  char *ok = "failed", *path;
  int n=0, i=0;
  int ret = (*orig_fchmod) (fd, mode);
  if(ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fchmod (%s, ", path);
  for(; chmod_mode[i].name && mode; i++) {
    if((chmod_mode[i].value & mode) == chmod_mode[i].value) {
      if(n>0)
	trace_print("|");
      trace_print(chmod_mode[i].name);
      mode &= ~chmod_mode[i].value;
      n++;
    }//if
  }//for
  trace_print(") ... %s\n ", ok);
  return ret;
}//trace_fchmod

int trace_fchown (int fd, uid_t owner, gid_t group)
{
  char *ok = "failed", *path;
  int ret = (*orig_fchown) (fd, owner, group);
  if(ret == 0)
    ok="done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fchown (%s,%d,%d) ... %s\n", path, owner, group, ok);
  return ret;
}//trace_fchown


int trace_fchown32 (int fd, uid_t owner, gid_t group)
{
  char *ok = "failed", *path;
  int ret = (*orig_fchown32) (fd, owner, group);
  if(ret == 0)
    ok="done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fchown32 (%s,%d,%d) ... %s\n", path, owner, group, ok);
  return ret;
}//trace_fchown32

int trace_fcntl (int fd, int cmd, ...)
{
  char *ok = "failed", *path;
  void *args = __builtin_apply_args();
  int *result = (int *) __builtin_apply ((void *) orig_fcntl, args, 64);
  if(*result > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fcntl (%s) ... %s\n", path, ok);
  __builtin_return (result);
}//trace_fcntl

int trace_fcntl64 (int fd, int cmd, ...)
{
  char *ok = "failed", *path;
  void *args = __builtin_apply_args();
  int *result = (int *) __builtin_apply ((void *) orig_fcntl64, args, 128);
  if(*result > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fcntl64 (%s) ... %s\n", path, ok);
  __builtin_return (result);
}//trace_fcntl64

int trace_flock (int fd, int operation)
{
  char *ok = "failed", *path;
  int ret = (*orig_flock) (fd, operation);
  if (ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: flock (%s) ... %s\n", path, ok);
  return ret;
}//trace_flock

int trace_fstat (int fd, struct stat *bufi)
{
  char *ok = "failed", *path;
  int ret = (*orig_fstat) (fd,bufi);
  if(ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fstat (%s) ... %s\n", path, ok);
  return ret;
}//trace_fstat

int trace_fstat64 (int fd, struct stat *bufi)
{
  char *ok = "failed", *path;
  int ret = (*orig_fstat64) (fd,bufi);
  if(ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fstat64 (%s) ... %s\n", path, ok);
  return ret;
}//trace_fstat64

int trace_fstatfs (int fd, struct statfs *bufi)
{
  char *ok = "failed", *path;
  int ret = (*orig_fstatfs) (fd,bufi);
  if(ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fstatfs (%s) ... %s\n", path, ok);
  return ret;
}//trace_fstatfs

int trace_fsync (int fd)
{
  char *ok = "failed", *path;
  int ret = (*orig_fsync) (fd);
  if(ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: fsync (%s) ... %s\n", path, ok);
  return ret;
}//trace_fsync

int trace_ftruncate (int fd, off_t length)
{
  char *ok = "failed", *path;
  int ret = (*orig_ftruncate) (fd, length);
  if(ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: ftruncate (%s, length=%ld) ... %s\n", path, length, ok);
  return ret;
}//trace_ftruncate

int trace_ftruncate64 (int fd, off_t length)
{
  char *ok = "failed", *path;
  int ret = (*orig_ftruncate64) (fd, length);
  if(ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: ftruncate64 (%s, length=%ld) ... %s\n", path, length, ok);
  return ret;
}//trace_ftruncate64

int trace_getdents (unsigned int fd, struct dirent *dirp, unsigned int count)
{
  char *ok = "failed", *path;
  int ret = (*orig_getdents) (fd, dirp, count);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: getdents (%s, count=%d) ... %s\n", path, count, ok);
  return ret;
}//trace_getdents

int trace_getdents64 (unsigned int fd, struct dirent *dirp, unsigned int count)
{
  char *ok = "failed", *path;
  int ret = (*orig_getdents64) (fd, dirp, count);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: getdents64 (%s, count=%d) ... %s\n", path, count, ok);
  return ret;
}//trace_getdents64

int trace_ioctl (int fd, int request, ...)
{
  char *ok = "failed", *path;
  void *args = __builtin_apply_args();
  int *result = (int *) __builtin_apply((void *) orig_ioctl, args, 128);
  if(*result > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: ioctl (%s) ... %s\n", path, ok);
  __builtin_return (result);
}//trace_ioctl

int trace_lchown (const char *path, uid_t owner, gid_t group)
{
  char *ok = "failed";
  int ret = (*orig_lchown) (path, owner, group);
  if(ret == 0)
    ok="done";
  trace_print("<1>vfs_trace: lchown (%s, owner=%d, group=%d) ... %s\n", path, owner, group, ok);
  return ret;
}//trace_lchown

int trace_lchown32 (const char *path, uid_t owner, gid_t group)
{
  char *ok = "failed";
  int ret = (*orig_lchown32) (path, owner, group);
  if(ret == 0)
    ok="done";
  trace_print("<1>vfs_trace: lchown32 (%s, owner=%d, group=%d) ... %s\n", path, owner, group, ok);
  return ret;
}//trace_lchown32

int trace_link (const char *oldpath, const char *newpath)
{
  char *ok = "failed";
  int ret = (*orig_link) (oldpath, newpath);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: link (%s, %s) ... %s\n", oldpath, newpath, ok);
  return ret;
}//trace_link

int trace__llseek (unsigned int fd, unsigned long offset_high, unsigned long offset_low, loff_t *result, unsigned int whence)
{
  char *ok = "failed", *path;
  int ret = (*orig__llseek) (fd, offset_high, offset_low, result, whence);
  if(ret == 0)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: _llseek (%s, %ld, %ld) ... %s\n", path, offset_high, offset_low, ok);
  return ret;
}//trace__llseek

off_t trace_lseek (int fd, off_t offset, int whence)
{
  char *ok = "failed", *path;
  off_t ret = (*orig_lseek) (fd, offset, whence);
  if(ret != (off_t) -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: lseek (%s) ... %s\n", path, ok);
  return ret;
}//trace_lseek

int trace_lstat (const char *path, struct stat *bufi)
{
  char *ok = "failed";
  int ret = (*orig_lstat) (path,bufi);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: lstat (%s) ... %s\n", path, ok);
  return ret;
}//trace_lstat

int trace_lstat64 (const char *path, struct stat *bufi)
{
  char *ok = "failed";
  int ret = (*orig_lstat64) (path,bufi);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: lstat64 (%s) ... %s\n", path, ok);
  return ret;
}//trace_lstat64

int trace_mkdir (const char *path, mode_t mode)
{
  char *ok = "failed";
  int ret = (*orig_mkdir) (path, mode);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: mkdir (%s) ... %s\n", path, ok);
  return ret;
}//trace_mkdir

int trace_mknod (const char *path, mode_t mode, dev_t dev)
{
  char *ok = "failed";
  int ret = (*orig_mknod) (path, mode, dev);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: mknod (%s) ... %s\n", path, ok);
  return ret;
}//trace_mknod

int trace_mount (const char *source, const char *target, const char *filesystemtype, unsigned long mountflags, const void *data)
{
  char *ok = "failed";
  int ret = (*orig_mount) (source, target, filesystemtype, mountflags, data);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: mount (%s, %s, %s) ... %s\n", source, target, filesystemtype, ok);
  return ret;
}//trace_mount

int trace_pipe (int fd[2])
{
  char *ok = "failed", *filedes0, *filedes1;
  int ret = (*orig_pipe) (fd);
  if(ret == 0)
    ok = "done";
  filedes0 = (char *) current->files->fd[fd[0]]->f_dentry->d_name.name;
  filedes1 = (char *) current->files->fd[fd[1]]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: pipe (%s,%s) ... %s\n", filedes0, filedes1, ok);
  return ret;
}//trace_pipe

int trace_pivot_root (const char *new_root, const char *put_old)
{
  char *ok = "failed";
  int ret = (*orig_pivot_root) (new_root, put_old);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: pivot_root (%s, %s) ... %s", new_root, put_old, ok);
  return ret;
}//trace_pivot_root

int trace_poll (struct pollfd *ufds, unsigned int nfds, int timeout)
{
  char *ok = "failed", *path;
  int ret = (*orig_poll) (ufds, nfds, timeout);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[ufds->fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: poll (%s) ... %s\n", path, ok);
  return ret;
}//trace_poll

ssize_t trace_pread (int fd, void *bufi, size_t count, off_t offset)
{
  char *ok = "failed", *path;
  ssize_t ret = (*orig_pread) (fd, bufi, count, offset);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: pread (%s, count=%d) ... %s\n", path, count, ok);
  return ret;
}//trace_pread

ssize_t trace_pwrite (int fd, const void *bufi, size_t count, off_t offset)
{
  char *ok = "failed", *path;
  ssize_t ret = (*orig_pwrite) (fd, bufi, count, offset);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: pwrite (%s, count=%d) ... %s\n", path, count, ok);
  return ret;
}//trace_pwrite

ssize_t trace_read (int fd, void *bufi, size_t count)
{
  char *ok = "failed", *path;
  ssize_t ret = (*orig_read) (fd, bufi, count);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: read (%s, count=%d) ... %s\n", path, count, ok);
  return ret;
}//trace_read

int trace_readdir (unsigned int fd, struct dirent *dirp, unsigned int count)
{
  char *ok = "failed", *path;
  int ret = (*orig_readdir) (fd, dirp, count);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: readdir (%s, count=%d) ... %s\n", path, count, ok);
  return ret;
}//trace_readdir

int trace_readlink (const char *path, char *bufi, size_t bufsiz)
{
  char *ok = "failed";
  int ret = (*orig_readlink) (path, bufi, bufsiz);
  if(ret > -1)
    ok = "done";
  trace_print("<1>vfs_trace: readlink (%s, bufsiz=%d) ... %s\n", path, bufsiz, ok);
  return ret;
}//trace_readlink

int trace_readv (int fd, const struct iovec *vector, size_t count)
{
  char *ok = "failed", *path;
  int ret = (*orig_readv) (fd, vector, count);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: readv (%s, count=%d) ... %s\n", path, count, ok);
  return ret;
}//trace_readv

int trace_rmdir (const char *path)
{
  char *ok = "failed";
  int ret = (*orig_rmdir) (path);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: rmdir (%s) ... %s\n", path, ok);
  return ret;
}//trace_rmdir

int trace_rename (const char *oldpath, const char *newpath)
{
  char *ok = "failed";
  int ret = (*orig_rename) (oldpath, newpath);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: rename (%s,%s) ... %s\n", oldpath, newpath, ok);
  return ret;
}//trace_rename

int trace_select (int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout)
{
  char *ok = "failed";
  int ret = (*orig_select) (n, readfds, writefds, exceptfds, timeout);
  if(ret > -1)
    ok = "done";
  trace_print("<1>vfs_trace: select (%d) ... %s\n", n, ok);
  return ret;
}//trace_select

int trace_stat (const char *path, struct stat *bufi)
{
  char *ok = "failed";
  int ret = (*orig_stat) (path,bufi);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: stat (%s) ... %s\n", path, ok);
  return ret;
}//trace_stat

int trace_stat64 (const char *path, struct stat *bufi)
{
  char *ok = "failed";
  int ret = (*orig_stat64) (path,bufi);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: stat64 (%s) ... %s\n", path, ok);
  return ret;
}//trace_stat64

int trace_statfs (const char *path, struct statfs *bufi)
{
  char *ok = "failed";
  int ret = (*orig_statfs) (path,bufi);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: statfs (%s) ... %s\n", path, ok);
  return ret;
}//trace_statfs

int trace_symlink (const char *oldpath, const char *newpath)
{
  char *ok = "failed";
  int ret = (*orig_symlink) (oldpath, newpath);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: symlink (%s, %s) ... %s\n", oldpath, newpath, ok);
  return ret;
}//trace_symlink

void trace_sync (void)
{
  (*orig_sync) ();
  trace_print("<1>vfs_trace: sync ... done\n");
}//trace_sync

int trace_sysfs (int option, ...)
{
  char *ok = "failed";
  void *args = __builtin_apply_args();
  int *result = (int *) __builtin_apply((void *) orig_ioctl, args, 128);
  if(*result > -1)
    ok = "done";
  trace_print("<1>vfs_trace: sysfs (%d) ... %s\n", option, ok);
  __builtin_return (result);
}//trace_sysfs

int trace_truncate (const char *path, off_t length)
{
  char *ok = "failed";
  int ret = (*orig_truncate) (path,length);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: truncate (%s, length=%ld) ... %s\n", path, length, ok);
  return ret;
}//trace_truncate

int trace_truncate64 (const char *path, off_t length)
{
  char *ok = "failed";
  int ret = (*orig_truncate64) (path,length);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: truncate64 (%s, length=%ld) ... %s\n", path, length, ok);
  return ret;
}//trace_truncate64

mode_t trace_umask (mode_t mask)
{
  trace_print("<1>vfs_trace: umask\n");
  return (*orig_umask) (mask);
}//trace_umask

int trace_umount (const char *target)
{
  char *ok = "failed";
  int ret = (*orig_umount) (target);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: umount (%s) ... %s\n", target, ok);
  return ret;
}//trace_umount

int trace_unlink (const char *path)
{
  char *ok = "failed";
  int ret = (*orig_unlink) (path);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: unlink (%s) ... %s\n", path, ok);
  return ret;
}//trace_unlink

int trace_uselib (const char *library)
{
  char *ok = "failed";
  int ret = (*orig_uselib) (library);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: uselib (%s) ... %s\n", library, ok);
  return ret;
}//trace_uselib

int trace_ustat (dev_t dev, struct ustat *ubuf)
{
  char *ok = "failed";
  int ret = (*orig_ustat) (dev, ubuf);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: ustat (%d) ... %s\n", dev, ok);
  return ret;
}//trace_ustat

int trace_utime (const char *path, struct utimbuf *bufi)
{
  char *ok = "failed";
  int ret = (*orig_utime) (path, bufi);
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: utime (%s) ... %s\n", path, ok);
  return ret;
}//trace_utime

int trace_vhangup (void)
{
  char *ok = "failed";
  int ret = (*orig_vhangup) ();
  if(ret == 0)
    ok = "done";
  trace_print("<1>vfs_trace: vhangup (void) ... %s\n", ok);
  return ret;
}//vhangup

ssize_t trace_write (int fd, const void *bufi, size_t count)
{
  char *ok = "failed", *path;
  ssize_t ret = (*orig_write) (fd, bufi, count);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: write (%s, count=%d) ... %s\n", path, count, ok);
  return ret;
}//trace_write

int trace_writev (int fd, const struct iovec *vector, size_t count)
{
  char *ok = "failed", *path;
  int ret = (*orig_writev) (fd, vector, count);
  if(ret > -1)
    ok = "done";
  path = (char *) current->files->fd[fd]->f_dentry->d_name.name;
  trace_print("<1>vfs_trace: writev (%s, count=%d) ... %s\n", path, count, ok);
  return ret;
}//trace_writev


/* ------------------------ main mathods -----------------------*/
int init_module(void)
{
  printk("<1>--------- Beginne with VFS tracing. ---------\n");

  //start = time(NULL);

  orig_access = sys_call_table[SYS_access];
  orig_bdflush = sys_call_table[SYS_bdflush];
  orig_chdir = sys_call_table[SYS_chdir];
  orig_chmod = sys_call_table[SYS_chmod];
  orig_chown = sys_call_table[SYS_chown];
  orig_chown32 = sys_call_table[SYS_chown32];
  orig_chroot = sys_call_table[SYS_chroot];
  orig_close = sys_call_table[SYS_close];
  orig_creat = sys_call_table[SYS_creat];
  orig_dup = sys_call_table[SYS_dup];
  orig_dup2 = sys_call_table[SYS_dup2];
  orig_execve = sys_call_table[SYS_execve];
  #ifdef _POSIX_SYNCHRONIZED_IO
    orig_fdatasync = sys_call_table[SYS_fdatasync];
  #endif
  orig_fchdir = sys_call_table[SYS_fchdir];
  orig_fchmod = sys_call_table[SYS_fchmod];
  orig_fchown = sys_call_table[SYS_fchown];
  orig_fchown32 = sys_call_table[SYS_fchown32]; 
  orig_fcntl = sys_call_table[SYS_fcntl];
  orig_fcntl64 = sys_call_table[SYS_fcntl64];
  orig_fsync = sys_call_table[SYS_fsync];
  orig_flock = sys_call_table[SYS_flock]; 
  orig_fstat = sys_call_table[SYS_fstat];
  orig_fstat64 = sys_call_table[SYS_fstat64];
  orig_fstatfs = sys_call_table[SYS_fstatfs];
  orig_ftruncate = sys_call_table[SYS_ftruncate];
  orig_ftruncate64 = sys_call_table[SYS_ftruncate64];
  orig_getdents = sys_call_table[SYS_getdents];
  orig_getdents64 = sys_call_table[SYS_getdents64];
  orig_ioctl = sys_call_table[SYS_ioctl];
  orig_lchown = sys_call_table[SYS_lchown];
  orig_lchown32 = sys_call_table[SYS_lchown32];
  orig_link = sys_call_table[SYS_link];
  orig__llseek = sys_call_table[SYS__llseek];
  orig_lseek = sys_call_table[SYS_lseek];
  orig_lstat = sys_call_table[SYS_lstat];
  orig_lstat64 = sys_call_table[SYS_lstat64];
  orig_mkdir = sys_call_table[SYS_mkdir];
  orig_mknod = sys_call_table[SYS_mknod];
  orig_mount = sys_call_table[SYS_mount];
  orig_open = sys_call_table[SYS_open];
  orig_pipe = sys_call_table[SYS_pipe];
  orig_pivot_root = sys_call_table[SYS_pivot_root];
  orig_poll = sys_call_table[SYS_poll];
  orig_pread = sys_call_table[SYS_pread];
  orig_pwrite = sys_call_table[SYS_pwrite];
  orig_read = sys_call_table[SYS_read];
  orig_readdir = sys_call_table[SYS_readdir];
  orig_readlink = sys_call_table[SYS_readlink];
  orig_readv = sys_call_table[SYS_readv];
  orig_rmdir = sys_call_table[SYS_rmdir];
  orig_rename = sys_call_table[SYS_rename];
  orig_select = sys_call_table[SYS_select];
  orig_stat = sys_call_table[SYS_stat];
  orig_stat64 = sys_call_table[SYS_stat64];
  orig_statfs = sys_call_table[SYS_statfs];
  orig_symlink = sys_call_table[SYS_symlink];
  orig_sync = sys_call_table[SYS_sync];
  orig_sysfs = sys_call_table[SYS_sysfs];
  orig_truncate = sys_call_table[SYS_truncate];
  orig_truncate64 = sys_call_table[SYS_truncate64];
  orig_umask = sys_call_table[SYS_umask];
  orig_umount = sys_call_table[SYS_umount];
  orig_unlink = sys_call_table[SYS_unlink];
  orig_uselib = sys_call_table[SYS_uselib];
  orig_ustat = sys_call_table[SYS_ustat];
  orig_utime = sys_call_table[SYS_utime];
  orig_vhangup = sys_call_table[SYS_vhangup];
  orig_write = sys_call_table[SYS_write];
  orig_writev = sys_call_table[SYS_writev];

  sys_call_table[SYS_access] = trace_access;
  sys_call_table[SYS_bdflush] = trace_bdflush;
  sys_call_table[SYS_chdir] = trace_chdir;
  sys_call_table[SYS_chmod] = trace_chmod;
  sys_call_table[SYS_chown] = trace_chown;
  sys_call_table[SYS_chown32] = trace_chown32;
  sys_call_table[SYS_chroot] = trace_chroot;
  sys_call_table[SYS_close] = trace_close;
  sys_call_table[SYS_creat] = trace_creat;
  sys_call_table[SYS_dup] = trace_dup;
  sys_call_table[SYS_dup2] = trace_dup2; 
  //sys_call_table[SYS_execve] = trace_execve;
  #ifdef _POSIX_SYNCHRONIZED_IO
  sys_call_table[SYS_fdatasync] = trace_fdatasync;
  #endif
  sys_call_table[SYS_fchdir] = trace_fchdir;
  sys_call_table[SYS_fchmod] = trace_fchmod;
  sys_call_table[SYS_fchown] = trace_fchown;
  sys_call_table[SYS_fchown32] = trace_fchown32;
  sys_call_table[SYS_fcntl] = trace_fcntl;
  //sys_call_table[SYS_fcntl64] = trace_fcntl64;//NULL pointer
  sys_call_table[SYS_flock] = trace_flock; 
  sys_call_table[SYS_fstat] = trace_fstat;

  //sys_call_table[SYS_fstat64] = trace_fstat64; //
  sys_call_table[SYS_fstatfs] = trace_fstatfs;
  sys_call_table[SYS_fsync] = trace_fsync;
  sys_call_table[SYS_ftruncate] = trace_ftruncate;
  sys_call_table[SYS_ftruncate64] = trace_ftruncate64;
  sys_call_table[SYS_getdents] = trace_getdents;
  sys_call_table[SYS_getdents64] = trace_getdents64;

  //sys_call_table[SYS_ioctl] = trace_ioctl; //wirf exception am Ende
  sys_call_table[SYS_lchown] = trace_lchown;
  sys_call_table[SYS_lchown32] = trace_lchown32;
  sys_call_table[SYS_link] = trace_link;
  sys_call_table[SYS__llseek] = trace__llseek;
  sys_call_table[SYS_lseek] = trace_lseek;
  sys_call_table[SYS_lstat] = trace_lstat;

  sys_call_table[SYS_lstat64] = trace_lstat64;
  sys_call_table[SYS_mkdir] = trace_mkdir;
  sys_call_table[SYS_mknod] = trace_mknod;
  sys_call_table[SYS_mount] = trace_mount;

  sys_call_table[SYS_open] = trace_open; 
  sys_call_table[SYS_pipe] = trace_pipe;
  sys_call_table[SYS_pivot_root] = trace_pivot_root;
  sys_call_table[SYS_poll] = trace_poll;
  sys_call_table[SYS_pread] = trace_pread;
  sys_call_table[SYS_pwrite] = trace_pwrite;
  sys_call_table[SYS_read] = trace_read;
  sys_call_table[SYS_readdir] = trace_readdir;
  sys_call_table[SYS_readlink] = trace_readlink;
  sys_call_table[SYS_readv] = trace_readv;
  sys_call_table[SYS_rmdir] = trace_rmdir;
  sys_call_table[SYS_rename] = trace_rename;
  sys_call_table[SYS_select] = trace_select;
  sys_call_table[SYS_stat] = trace_stat;
  sys_call_table[SYS_stat64] = trace_stat64;
  sys_call_table[SYS_statfs] = trace_statfs;
  sys_call_table[SYS_symlink] = trace_symlink;
  sys_call_table[SYS_sync] = trace_sync;
  sys_call_table[SYS_sysfs] = trace_sysfs;
  sys_call_table[SYS_truncate] = trace_truncate; 
  sys_call_table[SYS_truncate64] = trace_truncate64; 
  sys_call_table[SYS_umask] = trace_umask; 
  sys_call_table[SYS_umount] = trace_umount; 
  sys_call_table[SYS_unlink] = trace_unlink; 
  sys_call_table[SYS_uselib] = trace_uselib;
  sys_call_table[SYS_ustat] = trace_ustat;
  sys_call_table[SYS_utime] = trace_utime;
  sys_call_table[SYS_vhangup] = trace_vhangup;
  sys_call_table[SYS_write] = trace_write;
  sys_call_table[SYS_writev] = trace_writev;

  return 0;
}//init_module

void cleanup_module(void)
{

  printk("<1>vfs_trace: Cleaning up ... ");
  
  sys_call_table[SYS_access] = orig_access;
  sys_call_table[SYS_bdflush] = orig_bdflush;
  sys_call_table[SYS_chdir] = orig_chdir;
  sys_call_table[SYS_chmod] = orig_chmod;
  sys_call_table[SYS_chown] = orig_chown;
  sys_call_table[SYS_chown32] = orig_chown32;
  sys_call_table[SYS_chroot] = orig_chroot;
  sys_call_table[SYS_close] = orig_close;
  sys_call_table[SYS_creat] = orig_creat;
  sys_call_table[SYS_dup] = orig_dup;
  sys_call_table[SYS_dup2] = orig_dup2;
  #ifdef _POSIX_SYNCHRONIZED_IO
    sys_call_table[SYS_fdatasync] = orig_fdatasync;
  #endif
  sys_call_table[SYS_execve] = orig_execve;
  sys_call_table[SYS_fchdir] = orig_fchdir;
  sys_call_table[SYS_fchmod] = orig_fchmod;
  sys_call_table[SYS_fchown] = orig_fchown;
  sys_call_table[SYS_fchown32] = orig_fchown32;
  sys_call_table[SYS_fcntl] = orig_fcntl;
  sys_call_table[SYS_fcntl64] = orig_fcntl64;
  sys_call_table[SYS_flock] = orig_flock; 
  sys_call_table[SYS_fstat] = orig_fstat;
  sys_call_table[SYS_fstat64] = orig_fstat64;
  sys_call_table[SYS_fstatfs] = orig_fstatfs;
  sys_call_table[SYS_fsync] = orig_fsync;
  sys_call_table[SYS_ftruncate] = orig_ftruncate;
  sys_call_table[SYS_ftruncate64] = orig_ftruncate64;
  sys_call_table[SYS_getdents] = orig_getdents;
  sys_call_table[SYS_getdents64] = orig_getdents64;
  sys_call_table[SYS_ioctl] = orig_ioctl;
  sys_call_table[SYS_lchown] = orig_lchown;
  sys_call_table[SYS_lchown32] = orig_lchown32;
  sys_call_table[SYS_link] = orig_link;
  sys_call_table[SYS__llseek] = orig__llseek;
  sys_call_table[SYS_lseek] = orig_lseek;
  sys_call_table[SYS_lstat] = orig_lstat;
  sys_call_table[SYS_lstat64] = orig_lstat64;
  sys_call_table[SYS_mkdir] = orig_mkdir;
  sys_call_table[SYS_mknod] = orig_mknod;
  sys_call_table[SYS_mount] = orig_mount;
  sys_call_table[SYS_open] = orig_open;
  sys_call_table[SYS_pipe] = orig_pipe;
  sys_call_table[SYS_pivot_root] = orig_pivot_root;
  sys_call_table[SYS_poll] = orig_poll;
  sys_call_table[SYS_pread] = orig_pread;
  sys_call_table[SYS_pwrite] = orig_pwrite;
  sys_call_table[SYS_read] = orig_read;
  sys_call_table[SYS_readdir] = orig_readdir;
  sys_call_table[SYS_readlink] = orig_readlink;
  sys_call_table[SYS_readv] = orig_readv;
  sys_call_table[SYS_rmdir] = orig_rmdir;
  sys_call_table[SYS_rename] = orig_rename;
  sys_call_table[SYS_select] = orig_select;
  sys_call_table[SYS_stat] = orig_stat;
  sys_call_table[SYS_stat64] = orig_stat64;
  sys_call_table[SYS_statfs] = orig_statfs;
  sys_call_table[SYS_symlink] = orig_symlink;
  sys_call_table[SYS_sync] = orig_sync;
  sys_call_table[SYS_sysfs] = orig_sysfs;
  sys_call_table[SYS_truncate] = orig_truncate;
  sys_call_table[SYS_truncate64] = orig_truncate64;
  sys_call_table[SYS_umask] = orig_umask;
  sys_call_table[SYS_umount] = orig_umount;
  sys_call_table[SYS_unlink] = orig_unlink;
  sys_call_table[SYS_uselib] = orig_uselib;
  sys_call_table[SYS_ustat] = orig_ustat;
  sys_call_table[SYS_utime] = orig_utime;
  sys_call_table[SYS_vhangup] = orig_vhangup;
  sys_call_table[SYS_write] = orig_write;
  sys_call_table[SYS_writev] = orig_writev;

  printk("done.\n");
  printk("<1>--------- Ende with VFS tracing. ---------\n");
}//cleanup_module

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
